/*
 * Copyright 2023-2024 FalsePattern
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.falsepattern.zigbrains.lsp.client.languageserver.wrapper;

import com.falsepattern.zigbrains.common.util.ApplicationUtil;
import com.falsepattern.zigbrains.common.util.FileUtil;
import com.falsepattern.zigbrains.lsp.IntellijLanguageClient;
import com.falsepattern.zigbrains.lsp.client.DefaultLanguageClient;
import com.falsepattern.zigbrains.lsp.client.ServerWrapperBaseClientContext;
import com.falsepattern.zigbrains.lsp.client.languageserver.ServerOptions;
import com.falsepattern.zigbrains.lsp.client.languageserver.ServerStatus;
import com.falsepattern.zigbrains.lsp.client.languageserver.requestmanager.DefaultRequestManager;
import com.falsepattern.zigbrains.lsp.client.languageserver.requestmanager.RequestManager;
import com.falsepattern.zigbrains.lsp.client.languageserver.serverdefinition.LanguageServerDefinition;
import com.falsepattern.zigbrains.lsp.editor.EditorEventManager;
import com.falsepattern.zigbrains.lsp.editor.EditorEventManagerBase;
import com.falsepattern.zigbrains.lsp.extensions.LSPExtensionManager;
import com.falsepattern.zigbrains.lsp.listeners.DocumentListenerImpl;
import com.falsepattern.zigbrains.lsp.listeners.LSPCaretListenerImpl;
import com.falsepattern.zigbrains.lsp.requests.Timeout;
import com.falsepattern.zigbrains.lsp.requests.Timeouts;
import com.falsepattern.zigbrains.lsp.statusbar.LSPServerStatusWidget;
import com.falsepattern.zigbrains.lsp.statusbar.LSPServerStatusWidgetFactory;
import com.falsepattern.zigbrains.lsp.utils.FileUtils;
import com.falsepattern.zigbrains.lsp.utils.LSPException;
import com.google.gson.GsonBuilder;
import com.intellij.codeInsight.daemon.DaemonCodeAnalyzer;
import com.intellij.openapi.application.ApplicationInfo;
import com.intellij.openapi.diagnostic.Logger;
import com.intellij.openapi.editor.Editor;
import com.intellij.openapi.fileEditor.FileEditor;
import com.intellij.openapi.fileEditor.FileEditorManager;
import com.intellij.openapi.project.Project;
import com.intellij.openapi.project.ProjectUtil;
import com.intellij.openapi.ui.MessageType;
import com.intellij.openapi.ui.Messages;
import com.intellij.openapi.vfs.VirtualFile;
import com.intellij.psi.PsiDocumentManager;
import com.intellij.psi.PsiFile;
import com.intellij.remoteServer.util.CloudNotifier;
import com.intellij.util.PlatformIcons;
import lombok.val;
import org.apache.commons.lang3.tuple.ImmutablePair;
import org.apache.commons.lang3.tuple.Pair;
import org.eclipse.lsp4j.ClientCapabilities;
import org.eclipse.lsp4j.ClientInfo;
import org.eclipse.lsp4j.CodeActionCapabilities;
import org.eclipse.lsp4j.CodeActionKindCapabilities;
import org.eclipse.lsp4j.CodeActionLiteralSupportCapabilities;
import org.eclipse.lsp4j.CompletionCapabilities;
import org.eclipse.lsp4j.CompletionItemCapabilities;
import org.eclipse.lsp4j.DefinitionCapabilities;
import org.eclipse.lsp4j.DidChangeWatchedFilesCapabilities;
import org.eclipse.lsp4j.DocumentHighlightCapabilities;
import org.eclipse.lsp4j.ExecuteCommandCapabilities;
import org.eclipse.lsp4j.FoldingRangeCapabilities;
import org.eclipse.lsp4j.FoldingRangeKind;
import org.eclipse.lsp4j.FoldingRangeKindSupportCapabilities;
import org.eclipse.lsp4j.FoldingRangeSupportCapabilities;
import org.eclipse.lsp4j.FormattingCapabilities;
import org.eclipse.lsp4j.HoverCapabilities;
import org.eclipse.lsp4j.InitializeParams;
import org.eclipse.lsp4j.InitializeResult;
import org.eclipse.lsp4j.InitializedParams;
import org.eclipse.lsp4j.InlayHintCapabilities;
import org.eclipse.lsp4j.OnTypeFormattingCapabilities;
import org.eclipse.lsp4j.RangeFormattingCapabilities;
import org.eclipse.lsp4j.ReferencesCapabilities;
import org.eclipse.lsp4j.RenameCapabilities;
import org.eclipse.lsp4j.ServerCapabilities;
import org.eclipse.lsp4j.SignatureHelpCapabilities;
import org.eclipse.lsp4j.SymbolCapabilities;
import org.eclipse.lsp4j.SynchronizationCapabilities;
import org.eclipse.lsp4j.TextDocumentClientCapabilities;
import org.eclipse.lsp4j.TextDocumentSyncKind;
import org.eclipse.lsp4j.TextDocumentSyncOptions;
import org.eclipse.lsp4j.WorkspaceClientCapabilities;
import org.eclipse.lsp4j.WorkspaceEditCapabilities;
import org.eclipse.lsp4j.WorkspaceFolder;
import org.eclipse.lsp4j.jsonrpc.Launcher;
import org.eclipse.lsp4j.jsonrpc.ResponseErrorException;
import org.eclipse.lsp4j.jsonrpc.messages.Either;
import org.eclipse.lsp4j.jsonrpc.messages.Message;
import org.eclipse.lsp4j.jsonrpc.messages.ResponseErrorCode;
import org.eclipse.lsp4j.jsonrpc.messages.ResponseMessage;
import org.eclipse.lsp4j.services.LanguageClient;
import org.eclipse.lsp4j.services.LanguageServer;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.URI;
import java.net.URISyntaxException;
import java.nio.file.Path;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.Set;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;
import java.util.concurrent.atomic.AtomicReference;

import static com.falsepattern.zigbrains.lsp.client.languageserver.ServerStatus.INITIALIZED;
import static com.falsepattern.zigbrains.lsp.client.languageserver.ServerStatus.STARTED;
import static com.falsepattern.zigbrains.lsp.client.languageserver.ServerStatus.STARTING;
import static com.falsepattern.zigbrains.lsp.client.languageserver.ServerStatus.STOPPED;
import static com.falsepattern.zigbrains.lsp.client.languageserver.ServerStatus.STOPPING;

/**
 * The implementation of a LanguageServerWrapper (specific to a serverDefinition and a project)
 */
public class LanguageServerWrapper {
    public final LanguageServerDefinition serverDefinition;
    private final LSPExtensionManager extManager;
    private final Project project;
    private final HashSet<Editor> toConnect = new HashSet<>();
    @Nullable
    private final Path projectRootPath;
    private final HashSet<String> urisUnderLspControl = new HashSet<>();
    private final HashSet<Editor> connectedEditors = new HashSet<>();
    private final Map<String, Set<EditorEventManager>> uriToEditorManagers = new HashMap<>();
    private LanguageServer languageServer;
    private LanguageClient client;
    private RequestManager requestManager;
    private InitializeResult initializeResult;
    private Future<?> launcherFuture;
    private CompletableFuture<InitializeResult> initializeFuture;
    private int crashCount = 0;
    private volatile boolean alreadyShownTimeout = false;
    private volatile boolean alreadyShownCrash = false;
    private volatile ServerStatus status = STOPPED;
    private final AtomicReference<Thread> shutdownHook = new AtomicReference<>();
    private static final Map<Pair<String, String>, LanguageServerWrapper> uriToLanguageServerWrapper =
            new ConcurrentHashMap<>();
    private static final Map<Project, LanguageServerWrapper> projectToLanguageServerWrapper = new ConcurrentHashMap<>();
    private static final Logger LOG = Logger.getInstance(LanguageServerWrapper.class);
    private static final CloudNotifier notifier = new CloudNotifier("Language Server Protocol client");

    public LanguageServerWrapper(@NotNull LanguageServerDefinition serverDefinition, @NotNull Project project) {
        this(serverDefinition, project, null);
    }

    public LanguageServerWrapper(@NotNull LanguageServerDefinition serverDefinition, @NotNull Project project,
                                 @Nullable LSPExtensionManager extManager) {
        this.serverDefinition = serverDefinition;
        this.project = project;
        // We need to keep the project rootPath in addition to the project instance, since we cannot get the project
        // base path if the project is disposed.
        val projectDir = ProjectUtil.guessProjectDir(project);
        if (projectDir != null) {
            this.projectRootPath = projectDir.toNioPath();
        } else {
            this.projectRootPath = null;
        }
        this.extManager = extManager;
        projectToLanguageServerWrapper.put(project, this);
    }

    /**
     * @param uri     A file uri
     * @param project The related project
     * @return The wrapper for the given uri, or None
     */
    public static LanguageServerWrapper forUri(String uri, Project project) {
        return uriToLanguageServerWrapper.get(new ImmutablePair<>(uri, FileUtils.projectToUri(project)));
    }

    public static LanguageServerWrapper forVirtualFile(VirtualFile file, Project project) {
        return uriToLanguageServerWrapper.get(new ImmutablePair<>(FileUtil.URIFromVirtualFile(file), FileUtils.projectToUri(project)));
    }

    /**
     * @param editor An editor
     * @return The wrapper for the given editor, or None
     */
    public static LanguageServerWrapper forEditor(Editor editor) {
        return uriToLanguageServerWrapper.get(new ImmutablePair<>(FileUtils.editorToURIString(editor), FileUtils.editorToProjectFolderUri(editor)));
    }

    public static LanguageServerWrapper forProject(Project project) {
        return projectToLanguageServerWrapper.get(project);
    }

    public LanguageServerDefinition getServerDefinition() {
        return serverDefinition;
    }

    public Path getProjectRootPath() {
        return projectRootPath;
    }

    /**
     * @return if the server supports willSaveWaitUntil
     */
    public boolean isWillSaveWaitUntil() {
        return Optional.ofNullable(getServerCapabilities())
                .map(ServerCapabilities::getTextDocumentSync)
                .map(Either::getRight)
                .map(TextDocumentSyncOptions::getWillSaveWaitUntil)
                .orElse(false);
    }

    /**
     * Warning: this is a long running operation
     *
     * @return the languageServer capabilities, or null if initialization job didn't complete
     */
    @Nullable
    public ServerCapabilities getServerCapabilities() {
        if (initializeResult != null)
            return initializeResult.getCapabilities();
        else {
            try {
                start();
                if (initializeFuture != null) {
                    initializeFuture.get(Timeout.getTimeout(Timeouts.INIT), TimeUnit.MILLISECONDS);
                    notifySuccess(Timeouts.INIT);
                }
            } catch (TimeoutException e) {
                notifyFailure(Timeouts.INIT);
                String msg = String.format("%s \n is not initialized after %d seconds",
                                           serverDefinition.toString(), Timeout.getTimeout(Timeouts.INIT) / 1000);
                LOG.warn(msg, e);
                ApplicationUtil.invokeLater(() -> {
                    if (!alreadyShownTimeout) {
                        notifier.showMessage(msg, MessageType.WARNING);
                        alreadyShownTimeout = true;
                    }
                });
                stop(false);
            } catch (Exception e) {
                LOG.warn(e);
                stop(false);
            }
        }
        return initializeResult != null ? initializeResult.getCapabilities() : null;
    }

    public void notifyResult(Timeouts timeouts, boolean success) {
        getWidgets().forEach(widget -> widget.notifyResult(timeouts, success));
    }

    public void notifySuccess(Timeouts timeouts) {
        notifyResult(timeouts, true);
    }

    public void notifyFailure(Timeouts timeouts) {
        notifyResult(timeouts, false);
    }

    /**
     * Returns the EditorEventManager for a given uri
     * <p>
     * WARNING: actually a file can be present in multiple editors, this function just gives you one editor. use {@link #getEditorManagersFor(String)} instead
     * only use for document level events such as open, close, ...
     *
     * @param uri the URI as a string
     * @return the EditorEventManager (or null)
     */
    public EditorEventManager getEditorManagerFor(String uri) {
        FileEditor selectedEditor = FileEditorManager.getInstance(project).getSelectedEditor();

        if (selectedEditor == null) {
            return null;
        }
        VirtualFile currentOpenFile = selectedEditor.getFile();
        VirtualFile requestedFile = FileUtil.virtualFileFromURI(uri);
        if (currentOpenFile == null || requestedFile == null) {
            return null;
        }
        if (requestedFile.equals(currentOpenFile)) {
            return EditorEventManagerBase.forEditor((Editor) FileEditorManager.getInstance(project).getSelectedEditor());
        }
        if (uriToEditorManagers.containsKey(uri) && !uriToEditorManagers.get(uri).isEmpty()) {
            return (EditorEventManager) uriToEditorManagers.get(uri).toArray()[0];
        }

        return null;
    }

    public Set<EditorEventManager> getEditorManagersFor(String uri) {
        return uriToEditorManagers.get(uri);
    }

    /**
     * @return The request manager for this wrapper
     */
    public RequestManager getRequestManager() {
        return requestManager;
    }

    /**
     * @return whether the underlying connection to language languageServer is still active
     */
    public boolean isActive() {
        return launcherFuture != null && !launcherFuture.isDone() && !launcherFuture.isCancelled()
                && !alreadyShownTimeout && !alreadyShownCrash;
    }

    /**
     * Connects an editor to the languageServer
     *
     * @param editor the editor
     */
    public void connect(Editor editor) {
        if (editor == null) {
            LOG.warn("editor is null for " + serverDefinition);
            return;
        }
        if (!FileUtils.isEditorSupported(editor)) {
            LOG.debug("Editor hosts a unsupported file type by the LS library.");
            return;
        }

        String uri = FileUtils.editorToURIString(editor);
        if (connectedEditors.contains(editor)) {
            return;
        }
        ImmutablePair<String, String> key = new ImmutablePair<>(uri, FileUtils.editorToProjectFolderUri(editor));

        uriToLanguageServerWrapper.put(key, this);

        start();
        if (initializeFuture != null) {
            ServerCapabilities capabilities = getServerCapabilities();
            if (capabilities == null) {
                LOG.warn("Capabilities are null for " + serverDefinition);
                return;
            }

            initializeFuture.thenRun(() -> {
                if (connectedEditors.contains(editor)) {
                    return;
                }
                try {
                    Either<TextDocumentSyncKind, TextDocumentSyncOptions> syncOptions = capabilities.getTextDocumentSync();
                    if (syncOptions != null) {
                        //Todo - Implement
                        //  SelectionListenerImpl selectionListener = new SelectionListenerImpl();
                        DocumentListenerImpl documentListener = new DocumentListenerImpl();
                        LSPCaretListenerImpl caretListener = new LSPCaretListenerImpl();

                        ServerOptions serverOptions = new ServerOptions(capabilities);
                        EditorEventManager manager;
                        if (extManager != null) {
                            manager = extManager.getExtendedEditorEventManagerFor(editor, documentListener,
                                    caretListener, requestManager, serverOptions,
                                    this);
                            if (manager == null) {
                                manager = new EditorEventManager(editor, documentListener, caretListener,
                                        requestManager, serverOptions, this);
                            }
                        } else {
                            manager = new EditorEventManager(editor, documentListener, caretListener,
                                    requestManager, serverOptions, this);
                        }
                        // selectionListener.setManager(manager);
                        documentListener.setManager(manager);
                        caretListener.setManager(manager);
                        manager.registerListeners();
                        if (!urisUnderLspControl.contains(uri)) {
                            manager.documentEventManager.registerListeners();
                        }
                        urisUnderLspControl.add(uri);
                        connectedEditors.add(editor);
                        if (uriToEditorManagers.containsKey(uri)) {
                            uriToEditorManagers.get(uri).add(manager);
                        } else {
                            Set<EditorEventManager> set = new HashSet<>();
                            set.add(manager);
                            uriToEditorManagers.put(uri, set);
                            manager.documentOpened();
                        }
                        manager.initComplete();
                        LOG.info("Created a manager for " + uri);
                        synchronized (toConnect) {
                            toConnect.remove(editor);
                        }
                        for (Editor ed : new HashSet<>(toConnect)) {
                            connect(ed);
                        }
                        // Triggers annotators since this is the first editor which starts the LS
                        // and annotators are executed before LS is bootstrap to provide diagnostics.
                        ApplicationUtil.computableReadAction(() -> {
                            PsiFile psiFile = PsiDocumentManager.getInstance(project).getPsiFile(editor.getDocument());
                            if (psiFile != null) {
                                DaemonCodeAnalyzer.getInstance(project).restart(psiFile);
                            }
                            return null;
                        });
                    }
                } catch (Exception e) {
                    LOG.error(e);
                }
            });

        } else {
            synchronized (toConnect) {
                toConnect.add(editor);
            }
        }
    }

    /*
     * The shutdown request is sent from the client to the server. It asks the server to shut down, but to not exit \
     * (otherwise the response might not be delivered correctly to the client).
     * Only if the exit flag is true, particular server instance will exit.
     */
    public void stop(boolean exit) {
        if (this.status == STOPPED || this.status == STOPPING) {
            return;
        }
        setStatus(STOPPING);

        if (initializeFuture != null) {
            initializeFuture.cancel(true);
        }

        try {
            if (languageServer != null) {
                CompletableFuture<Object> shutdown = languageServer.shutdown();
                shutdown.get(Timeout.getTimeout(Timeouts.SHUTDOWN), TimeUnit.MILLISECONDS);
                notifySuccess(Timeouts.SHUTDOWN);
                if (exit) {
                    languageServer.exit();
                }
            }
        } catch (Exception e) {
            // most likely closed externally.
            notifyFailure(Timeouts.SHUTDOWN);
            LOG.warn("exception occured while trying to shut down", e);
        } finally {
            if (launcherFuture != null) {
                launcherFuture.cancel(true);
            }
            if (serverDefinition != null) {
                serverDefinition.stop(projectRootPath != null ? projectRootPath.toString() : null);
            }
            for (Editor ed : new HashSet<>(connectedEditors)) {
                disconnect(ed);
            }

            // sadly this whole editor closing stuff runs asynchronously, so we cannot be sure the state is really clean here...
            // therefore clear the mapping from here as it should be empty by now.
            uriToEditorManagers.clear();
            urisUnderLspControl.clear();
            launcherFuture = null;
            initializeResult = null;
            initializeFuture = null;
            languageServer = null;
            setStatus(STOPPED);
        }
    }

    /**
     * Checks if the wrapper is already connected to the document at the given path.
     *
     * @param location file location
     * @return True if the given file is connected.
     */
    public boolean isConnectedTo(String location) {
        return urisUnderLspControl.contains(location);
    }

    /**
     * @return the LanguageServer
     */
    @Nullable
    public LanguageServer getServer() {
        start();
        if (initializeFuture != null && !initializeFuture.isDone()) {
            initializeFuture.join();
        }
        return languageServer;
    }

    /**
     * Starts the LanguageServer
     */
    public void start() {
        if (status == STOPPED && !alreadyShownCrash && !alreadyShownTimeout) {
            setStatus(STARTING);
            try {
                Pair<InputStream, OutputStream> streams = serverDefinition.start(projectRootPath != null ? projectRootPath.toString() : null);
                InputStream inputStream = streams.getKey();
                OutputStream outputStream = streams.getValue();
                InitializeParams initParams = getInitParams();
                ExecutorService executorService = Executors.newCachedThreadPool();
                MessageHandler messageHandler = new MessageHandler(serverDefinition.getServerListener(), () -> getStatus() != STOPPED);
                if (extManager != null && extManager.getExtendedServerInterface() != null) {
                    Class<? extends LanguageServer> remoteServerInterFace = extManager.getExtendedServerInterface();
                    client = extManager.getExtendedClientFor(new ServerWrapperBaseClientContext(this));

                    val launcher = new Launcher.Builder<LanguageServer>()
                                           .setLocalService(client)
                                           .setRemoteInterface(remoteServerInterFace)
                                           .setInput(inputStream)
                                           .setOutput(outputStream)
                                           .setExecutorService(executorService)
                                           .wrapMessages(messageHandler)
                                           .configureGson(GsonBuilder::disableHtmlEscaping)
                                           .create();
                    languageServer = launcher.getRemoteProxy();
                    launcherFuture = launcher.startListening();
                } else {
                    client = new DefaultLanguageClient(new ServerWrapperBaseClientContext(this));
                    val launcher = new Launcher.Builder<LanguageServer>()
                                           .setLocalService(client)
                                           .setRemoteInterface(LanguageServer.class)
                                           .setInput(inputStream)
                                           .setOutput(outputStream)
                                           .setExecutorService(executorService)
                                           .wrapMessages(messageHandler)
                                           .configureGson(GsonBuilder::disableHtmlEscaping)
                                           .create();
                    languageServer = launcher.getRemoteProxy();
                    launcherFuture = launcher.startListening();
                }
                messageHandler.setLanguageServer(languageServer);

                initializeFuture = languageServer.initialize(initParams).thenApply(res -> {
                    initializeResult = res;
                    LOG.info("Got initializeResult for " + serverDefinition + " ; " + projectRootPath);
                    if (extManager != null) {
                        requestManager = extManager.getExtendedRequestManagerFor(this, languageServer, client, res.getCapabilities());
                        if (requestManager == null) {
                            requestManager = new DefaultRequestManager(this, languageServer, client, res.getCapabilities());
                        }
                    } else {
                        requestManager = new DefaultRequestManager(this, languageServer, client, res.getCapabilities());
                    }
                    setStatus(STARTED);
                    // send the initialized message since some language servers depends on this message
                    requestManager.initialized(new InitializedParams());
                    setStatus(INITIALIZED);
                    return res;
                });
            } catch (LSPException | IOException | URISyntaxException e) {
                LOG.warn(e);
                ApplicationUtil.invokeLater(() ->
                        notifier.showMessage(String.format("Can't start server due to %s", e.getMessage()),
                                MessageType.WARNING));
                removeServerWrapper();
            }
        }
    }

    private InitializeParams getInitParams() throws URISyntaxException {
        InitializeParams initParams = new InitializeParams();
        String projectRootUri = FileUtil.pathToUri(projectRootPath);
        if (projectRootUri.endsWith("/")) {
            projectRootUri = projectRootUri.substring(0, projectRootUri.length() - 1);
        }
        WorkspaceFolder workspaceFolder = new WorkspaceFolder(projectRootUri, this.project.getName());
        initParams.setWorkspaceFolders(Collections.singletonList(workspaceFolder));
        initParams.setProcessId((int) ProcessHandle.current().pid());
        initParams.setLocale("en-us");

        // workspace capabilities
        WorkspaceClientCapabilities workspaceClientCapabilities = new WorkspaceClientCapabilities();
        workspaceClientCapabilities.setApplyEdit(true);
        workspaceClientCapabilities.setDidChangeWatchedFiles(new DidChangeWatchedFilesCapabilities());
        workspaceClientCapabilities.setExecuteCommand(new ExecuteCommandCapabilities());
        workspaceClientCapabilities.setWorkspaceEdit(new WorkspaceEditCapabilities());
        workspaceClientCapabilities.setSymbol(new SymbolCapabilities());
        workspaceClientCapabilities.setWorkspaceFolders(true);
        workspaceClientCapabilities.setConfiguration(false);

        // text document capabilities
        TextDocumentClientCapabilities textDocumentClientCapabilities = new TextDocumentClientCapabilities();
        textDocumentClientCapabilities.setCodeAction(new CodeActionCapabilities());
        textDocumentClientCapabilities.getCodeAction().setCodeActionLiteralSupport(new CodeActionLiteralSupportCapabilities(new CodeActionKindCapabilities()));
        textDocumentClientCapabilities.setCompletion(new CompletionCapabilities(new CompletionItemCapabilities(true)));
        textDocumentClientCapabilities.setDefinition(new DefinitionCapabilities());
        textDocumentClientCapabilities.setDocumentHighlight(new DocumentHighlightCapabilities());
        textDocumentClientCapabilities.setFormatting(new FormattingCapabilities());
        textDocumentClientCapabilities.setHover(new HoverCapabilities());
        textDocumentClientCapabilities.setInlayHint(new InlayHintCapabilities());
        textDocumentClientCapabilities.setOnTypeFormatting(new OnTypeFormattingCapabilities());
        textDocumentClientCapabilities.setRangeFormatting(new RangeFormattingCapabilities());
        textDocumentClientCapabilities.setReferences(new ReferencesCapabilities());
        textDocumentClientCapabilities.setRename(new RenameCapabilities());
        textDocumentClientCapabilities.setSignatureHelp(new SignatureHelpCapabilities());
        textDocumentClientCapabilities.setSynchronization(new SynchronizationCapabilities(true, true, true));

        FoldingRangeCapabilities foldingRangeCapabilities = new FoldingRangeCapabilities();
        foldingRangeCapabilities.setFoldingRangeKind(new FoldingRangeKindSupportCapabilities(List.of(
                FoldingRangeKind.Comment, FoldingRangeKind.Imports, FoldingRangeKind.Region)));
        foldingRangeCapabilities.setFoldingRange(new FoldingRangeSupportCapabilities(true));
        textDocumentClientCapabilities.setFoldingRange(foldingRangeCapabilities);

        initParams.setCapabilities(
                new ClientCapabilities(workspaceClientCapabilities, textDocumentClientCapabilities, null));
        initParams.setClientInfo(new ClientInfo(ApplicationInfo.getInstance().getVersionName(), ApplicationInfo.getInstance().getFullVersion()));

        serverDefinition.customizeInitializeParams(initParams);
        return initParams;
    }

    public void logMessage(Message message) {
        if (message instanceof ResponseMessage) {
            ResponseMessage responseMessage = (ResponseMessage) message;
            if (responseMessage.getError() != null && (responseMessage.getId()
                    .equals(Integer.toString(ResponseErrorCode.RequestCancelled.getValue())))) {
                LOG.error(new ResponseErrorException(responseMessage.getError()));
            }
        }
    }

    public Project getProject() {
        return project;
    }

    public ServerStatus getStatus() {
        return status;
    }

    private void setStatus(ServerStatus status) {
        this.status = status;
        getWidgets().forEach(widget -> widget.setStatus(status));
        synchronized (shutdownHook) {
            if ((status == STARTED || status == INITIALIZED) && shutdownHook.get() == null) {
                shutdownHook.set(new Thread(() -> {
                    shutdownHook.set(null);
                    stop(true);
                }));
                Runtime.getRuntime().addShutdownHook(shutdownHook.get());
            } else if (status == STOPPED && shutdownHook.get() != null) {
                try {
                    Runtime.getRuntime().removeShutdownHook(shutdownHook.get());
                } catch (IllegalStateException ignored) {} //Shouldn't happen
                shutdownHook.set(null);
            }
        }
    }

    public void crashed(Exception e) {
        crashCount += 1;
        if (crashCount <= 3) {
            reconnect();
        } else {
            ApplicationUtil.invokeLater(() -> {
                if (alreadyShownCrash) {
                    reconnect();
                } else {
                    int response = Messages.showYesNoDialog(String.format(
                                    "LanguageServer for definition %s, project %s keeps crashing due to \n%s\n"
                                    , serverDefinition.toString(), project.getName(), e.getMessage()),
                            "Language Server Client Warning", "Keep Connected", "Disconnect", PlatformIcons.CHECK_ICON);
                    if (response == Messages.NO) {
                        int confirm = Messages.showYesNoDialog("All the language server based plugin features will be disabled.\n" +
                                "Do you wish to continue?", "", PlatformIcons.WARNING_INTRODUCTION_ICON);
                        if (confirm == Messages.YES) {
                            // Disconnects from the language server.
                            stop(true);
                        } else {
                            reconnect();
                        }
                    } else {
                        reconnect();
                    }
                }
                alreadyShownCrash = true;
                crashCount = 0;
            });
        }
    }

    private void reconnect() {
        // Need to copy by value since connected editors gets cleared during 'stop()' invocation.
        final Set<String> connected = new HashSet<>(urisUnderLspControl);
        stop(true);
        for (String uri : connected) {
            connect(uri);
        }
    }

    public List<String> getConnectedFiles() {
        List<String> connected = new ArrayList<>();
        urisUnderLspControl.forEach(s -> {
            try {
                connected.add(new URI(FileUtil.sanitizeURI(s)).toString());
            } catch (URISyntaxException e) {
                LOG.warn(e);
            }
        });
        return connected;
    }

    /**
     * Disconnects an editor from the LanguageServer
     *
     * @param editor The editor
     */
    public void disconnect(Editor editor) {
        EditorEventManager manager = EditorEventManagerBase.forEditor(editor);
        connectedEditors.remove(editor);
        if (manager != null) {
            manager.removeListeners();
            String uri = FileUtils.editorToURIString(editor);
            Set<EditorEventManager> set = uriToEditorManagers.get(uri);
            if (set != null) {
                set.remove(manager);
                if (set.isEmpty()) {
                    manager.documentClosed();
                    manager.documentEventManager.removeListeners();

                    uriToEditorManagers.remove(uri);
                    urisUnderLspControl.remove(uri);
                    uriToLanguageServerWrapper.remove(new ImmutablePair<>(uri, FileUtils.editorToProjectFolderUri(editor)));
                }
            }
        }

        if (connectedEditors.isEmpty()) {
            stop(true);

            getWidgets().forEach(widget -> widget.setStatus(ServerStatus.NONEXISTENT));
        }
    }

    /**
     * Disconnects an editor from the LanguageServer
     * <p>
     * WARNING: only use this method if you have no editor instance and you restart all connections to the language server for all open editors
     * prefer using disconnect(editor)
     *
     * @param uri        The file uri
     * @param projectUri The project root uri
     */
    public void disconnect(String uri, String projectUri) {
        uriToLanguageServerWrapper.remove(new ImmutablePair<>(FileUtil.sanitizeURI(uri), FileUtil.sanitizeURI(projectUri)));

        Set<EditorEventManager> managers = uriToEditorManagers.get(uri);
        if (managers == null) {
            return;
        }
        for (EditorEventManager manager : managers) {
            manager.removeListeners();
            manager.documentEventManager.removeListeners();
            connectedEditors.remove(manager.editor);
            Set<EditorEventManager> editorEventManagers = uriToEditorManagers.get(uri);
            if (editorEventManagers != null) {
                editorEventManagers.remove(manager);
                if (editorEventManagers.isEmpty()) {
                    uriToEditorManagers.remove(uri);
                    manager.documentClosed();
                }
            }
            urisUnderLspControl.remove(uri);
            uriToLanguageServerWrapper.remove(new ImmutablePair<>(FileUtil.sanitizeURI(uri), FileUtil.sanitizeURI(projectUri)));
        }
        if (connectedEditors.isEmpty()) {
            stop(true);

            getWidgets().forEach(widget -> widget.setStatus(ServerStatus.NONEXISTENT));
        }
    }

    public void removeServerWrapper() {
        stop(true);
        IntellijLanguageClient.removeWrapper(this);
    }

    private void connect(String uri) {
        List<Editor> editors = FileUtils.getAllOpenedEditorsForUri(project, uri);

        for (Editor editor : editors) {
            connect(editor);
        }
    }

    /**
     * Is the language server in a state where it can be restartable. Normally language server is
     * restartable if it has timeout or has a startup error.
     */
    public boolean isRestartable() {
        return status == STOPPED && (alreadyShownTimeout || alreadyShownCrash);
    }

    /**
     * Reset language server wrapper state so it can be started again if it was failed earlier.
     */
    public void restart() {
        ApplicationUtil.pool(() -> {
            if (isRestartable()) {
                alreadyShownCrash = false;
                alreadyShownTimeout = false;
            } else {
                stop(true);
            }
            FileUtils.reloadEditors(project);
        });
    }

    private List<LSPServerStatusWidget> getWidgets() {
        return Optional.ofNullable(project.getUserData(LSPServerStatusWidgetFactory.LSP_WIDGETS)).orElse(List.of());
    }

    /**
     * Returns the extension manager associated with this language server wrapper.
     *
     * @return The result can be null if there is not extension manager defined.
     */
    @Nullable
    public final LSPExtensionManager getExtensionManager() {
        return extManager;
    }
}

